@@ -115,7 +115,7 @@ int WavpackGetMode (WavpackContext *wpc)
 // and DSD files. Except for indicating the presence of DSD data, these
 // bits are meant to simply indicate the format of the data in the original
 // source file and do NOT indicate how the library will return the data to
-// the appication (which is always the same). This means that in general an
+// the application (which is always the same). This means that in general an
 // application that simply wants to play or process the audio data need not
 // be concerned about these. If the file is DSD audio, then either of the
 // QMDOE_DSD_LSB_FIRST or QMODE_DSD_MSB_FIRST bits will be set (but the
@@ -311,7 +311,7 @@ uint32_t WavpackGetChannelLayout (WavpackContext *wpc, unsigned char *reorder)
 // This function provides the identities of ALL the channels in the file, including the
 // standard Microsoft channels (which come first, in order, and are numbered 1-18) and also
 // any non-Microsoft channels (which can be in any order and have values from 33-254). The
-// value 0x00 is invalid and 0xFF indicates an "unknown" or "unnassigned" channel. The
+// value 0x00 is invalid and 0xFF indicates an "unknown" or "unassigned" channel. The
 // string is NULL terminated so the caller must supply enough space for the number
 // of channels indicated by WavpackGetNumChannels(), plus one.
 //
@@ -346,6 +346,31 @@ void WavpackGetChannelIdentities (WavpackContext *wpc, unsigned char *identities
     *identities = 0;
 }
 
+#ifdef ENABLE_THREADS
+
+static void worker_threads_destroy (WavpackContext *wpc)
+{
+    if (wpc->workers) {
+        int i;
+
+        for (i = 0; i < wpc->num_workers; ++i) {
+            wp_mutex_obtain (wpc->mutex);
+            wpc->workers [i].state = Quit;
+            wp_condvar_signal (wpc->workers [i].worker_cond);
+            wp_mutex_release (wpc->mutex);
+            wp_thread_join (wpc->workers [i].thread);
+            wp_thread_delete (wpc->workers [i].thread);
+            wp_condvar_delete (wpc->workers [i].worker_cond);
+        }
+
+        free (wpc->workers);
+        wpc->workers = NULL;
+        wp_mutex_delete (wpc->mutex);
+    }
+}
+
+#endif
+
 // For local use only. Install a callback to be executed when WavpackCloseFile() is called,
 // usually used to dump some statistics accumulated during encode or decode.
 
@@ -409,6 +434,10 @@ WavpackContext *WavpackCloseFile (WavpackContext *wpc)
         decimate_dsd_destroy (wpc->decimation_context);
 #endif
 
+#ifdef ENABLE_THREADS
+    worker_threads_destroy (wpc);
+#endif
+
     free (wpc);
 
     return NULL;
@@ -475,11 +504,13 @@ int WavpackGetChannelMask (WavpackContext *wpc)
 // Return the normalization value for floating point data (valid only
 // if floating point data is present). A value of 127 indicates that
 // the floating point range is +/- 1.0. Higher values indicate a
-// larger floating point range.
+// larger floating point range. Note that if the client specified
+// OPEN_NORMALIZE we return the normalized value (i.e., 127 + offset)
+// rather than what's in the file (which isn't really relevant).
 
 int WavpackGetFloatNormExp (WavpackContext *wpc)
 {
-    return wpc->config.float_norm_exp;
+    return (wpc->open_flags & OPEN_NORMALIZE) ? 127 + wpc->norm_offset : wpc->config.float_norm_exp;
 }
 
 // Returns the actual number of valid bits per sample contained in the
@@ -528,29 +559,7 @@ void free_streams (WavpackContext *wpc)
     int si = wpc->num_streams;
 
     while (si--) {
-        if (wpc->streams [si]->blockbuff) {
-            free (wpc->streams [si]->blockbuff);
-            wpc->streams [si]->blockbuff = NULL;
-        }
-
-        if (wpc->streams [si]->block2buff) {
-            free (wpc->streams [si]->block2buff);
-            wpc->streams [si]->block2buff = NULL;
-        }
-
-        if (wpc->streams [si]->sample_buffer) {
-            free (wpc->streams [si]->sample_buffer);
-            wpc->streams [si]->sample_buffer = NULL;
-        }
-
-        if (wpc->streams [si]->dc.shaping_data) {
-            free (wpc->streams [si]->dc.shaping_data);
-            wpc->streams [si]->dc.shaping_data = NULL;
-        }
-
-#ifdef ENABLE_DSD
-        free_dsd_tables (wpc->streams [si]);
-#endif
+        free_single_stream (wpc->streams [si]);
 
         if (si) {
             wpc->num_streams--;
@@ -558,10 +567,44 @@ void free_streams (WavpackContext *wpc)
             wpc->streams [si] = NULL;
         }
     }
-
-    wpc->current_stream = 0;
 }
 
+// Free all resources associated with the specified stream
+
+void free_single_stream (WavpackStream *wps)
+{
+    if (wps->blockbuff) {
+        free (wps->blockbuff);
+        wps->blockbuff = NULL;
+    }
+
+    if (wps->block2buff) {
+        free (wps->block2buff);
+        wps->block2buff = NULL;
+    }
+
+    if (wps->sample_buffer) {
+        free (wps->sample_buffer);
+        wps->sample_buffer = NULL;
+    }
+
+    if (wps->pre_sample_buffer) {
+        free (wps->pre_sample_buffer);
+        wps->sample_buffer = NULL;
+    }
+
+    if (wps->dc.shaping_data) {
+        free (wps->dc.shaping_data);
+        wps->dc.shaping_data = NULL;
+    }
+
+#ifdef ENABLE_DSD
+    free_dsd_tables (wps);
+#endif
+}
+
+// Free all DSD-related resources associated with the specified stream
+
 void free_dsd_tables (WavpackStream *wps)
 {
     if (wps->dsd.probabilities) {
@@ -615,26 +658,28 @@ void WavpackFloatNormalize (int32_t *values, int32_t num_values, int delta_exp)
 void WavpackLittleEndianToNative (void *data, char *format)
 {
     unsigned char *cp = (unsigned char *) data;
-    int64_t temp;
+    int64_t temp64;
+    int32_t temp32;
+    int16_t temp16;
 
     while (*format) {
         switch (*format) {
             case 'D':
-                temp = cp [0] + ((int64_t) cp [1] << 8) + ((int64_t) cp [2] << 16) + ((int64_t) cp [3] << 24) +
+                temp64 = cp [0] + ((int64_t) cp [1] << 8) + ((int64_t) cp [2] << 16) + ((int64_t) cp [3] << 24) +
                     ((int64_t) cp [4] << 32) + ((int64_t) cp [5] << 40) + ((int64_t) cp [6] << 48) + ((uint64_t) cp [7] << 56);
-                * (int64_t *) cp = temp;
+                memcpy (cp, &temp64, 8);
                 cp += 8;
                 break;
 
             case 'L':
-                temp = cp [0] + ((int32_t) cp [1] << 8) + ((int32_t) cp [2] << 16) + ((int64_t) cp [3] << 24);
-                * (int32_t *) cp = (int32_t) temp;
+                temp32 = cp [0] + ((int32_t) cp [1] << 8) + ((int32_t) cp [2] << 16) + ((int64_t) cp [3] << 24);
+                memcpy (cp, &temp32, 4);
                 cp += 4;
                 break;
 
             case 'S':
-                temp = cp [0] + (cp [1] << 8);
-                * (int16_t *) cp = (int16_t) temp;
+                temp16 = cp [0] + (cp [1] << 8);
+                memcpy (cp, &temp16, 2);
                 cp += 2;
                 break;
 
@@ -652,34 +697,36 @@ void WavpackLittleEndianToNative (void *data, char *format)
 void WavpackNativeToLittleEndian (void *data, char *format)
 {
     unsigned char *cp = (unsigned char *) data;
-    int64_t temp;
+    int64_t temp64;
+    int32_t temp32;
+    int16_t temp16;
 
     while (*format) {
         switch (*format) {
             case 'D':
-                temp = * (int64_t *) cp;
-                *cp++ = (unsigned char) temp;
-                *cp++ = (unsigned char) (temp >> 8);
-                *cp++ = (unsigned char) (temp >> 16);
-                *cp++ = (unsigned char) (temp >> 24);
-                *cp++ = (unsigned char) (temp >> 32);
-                *cp++ = (unsigned char) (temp >> 40);
-                *cp++ = (unsigned char) (temp >> 48);
-                *cp++ = (unsigned char) (temp >> 56);
+                memcpy (&temp64, cp, sizeof (temp64));
+                *cp++ = (unsigned char) temp64;
+                *cp++ = (unsigned char) (temp64 >> 8);
+                *cp++ = (unsigned char) (temp64 >> 16);
+                *cp++ = (unsigned char) (temp64 >> 24);
+                *cp++ = (unsigned char) (temp64 >> 32);
+                *cp++ = (unsigned char) (temp64 >> 40);
+                *cp++ = (unsigned char) (temp64 >> 48);
+                *cp++ = (unsigned char) (temp64 >> 56);
                 break;
 
             case 'L':
-                temp = * (int32_t *) cp;
-                *cp++ = (unsigned char) temp;
-                *cp++ = (unsigned char) (temp >> 8);
-                *cp++ = (unsigned char) (temp >> 16);
-                *cp++ = (unsigned char) (temp >> 24);
+                memcpy (&temp32, cp, sizeof (temp32));
+                *cp++ = (unsigned char) temp32;
+                *cp++ = (unsigned char) (temp32 >> 8);
+                *cp++ = (unsigned char) (temp32 >> 16);
+                *cp++ = (unsigned char) (temp32 >> 24);
                 break;
 
             case 'S':
-                temp = * (int16_t *) cp;
-                *cp++ = (unsigned char) temp;
-                *cp++ = (unsigned char) (temp >> 8);
+                memcpy (&temp16, cp, sizeof (temp16));
+                *cp++ = (unsigned char) temp16;
+                *cp++ = (unsigned char) (temp16 >> 8);
                 break;
 
             default:
@@ -696,26 +743,28 @@ void WavpackNativeToLittleEndian (void *data, char *format)
 void WavpackBigEndianToNative (void *data, char *format)
 {
     unsigned char *cp = (unsigned char *) data;
-    int64_t temp;
+    int64_t temp64;
+    int32_t temp32;
+    int16_t temp16;
 
     while (*format) {
         switch (*format) {
             case 'D':
-                temp = cp [7] + ((int64_t) cp [6] << 8) + ((int64_t) cp [5] << 16) + ((int64_t) cp [4] << 24) +
+                temp64 = cp [7] + ((int64_t) cp [6] << 8) + ((int64_t) cp [5] << 16) + ((int64_t) cp [4] << 24) +
                     ((int64_t) cp [3] << 32) + ((int64_t) cp [2] << 40) + ((int64_t) cp [1] << 48) + ((uint64_t) cp [0] << 56);
-                * (int64_t *) cp = temp;
+                memcpy (cp, &temp64, 8);
                 cp += 8;
                 break;
 
             case 'L':
-                temp = cp [3] + ((int32_t) cp [2] << 8) + ((int32_t) cp [1] << 16) + ((int64_t) cp [0] << 24);
-                * (int32_t *) cp = (int32_t) temp;
+                temp32 = cp [3] + ((int32_t) cp [2] << 8) + ((int32_t) cp [1] << 16) + ((int64_t) cp [0] << 24);
+                memcpy (cp, &temp32, 4);
                 cp += 4;
                 break;
 
             case 'S':
-                temp = cp [1] + (cp [0] << 8);
-                * (int16_t *) cp = (int16_t) temp;
+                temp16 = cp [1] + (cp [0] << 8);
+                memcpy (cp, &temp16, 2);
                 cp += 2;
                 break;
 
@@ -733,34 +782,36 @@ void WavpackBigEndianToNative (void *data, char *format)
 void WavpackNativeToBigEndian (void *data, char *format)
 {
     unsigned char *cp = (unsigned char *) data;
-    int64_t temp;
+    int64_t temp64;
+    int32_t temp32;
+    int16_t temp16;
 
     while (*format) {
         switch (*format) {
             case 'D':
-                temp = * (int64_t *) cp;
-                *cp++ = (unsigned char) (temp >> 56);
-                *cp++ = (unsigned char) (temp >> 48);
-                *cp++ = (unsigned char) (temp >> 40);
-                *cp++ = (unsigned char) (temp >> 32);
-                *cp++ = (unsigned char) (temp >> 24);
-                *cp++ = (unsigned char) (temp >> 16);
-                *cp++ = (unsigned char) (temp >> 8);
-                *cp++ = (unsigned char) temp;
+                memcpy (&temp64, cp, sizeof (temp64));
+                *cp++ = (unsigned char) (temp64 >> 56);
+                *cp++ = (unsigned char) (temp64 >> 48);
+                *cp++ = (unsigned char) (temp64 >> 40);
+                *cp++ = (unsigned char) (temp64 >> 32);
+                *cp++ = (unsigned char) (temp64 >> 24);
+                *cp++ = (unsigned char) (temp64 >> 16);
+                *cp++ = (unsigned char) (temp64 >> 8);
+                *cp++ = (unsigned char) temp64;
                 break;
 
             case 'L':
-                temp = * (int32_t *) cp;
-                *cp++ = (unsigned char) (temp >> 24);
-                *cp++ = (unsigned char) (temp >> 16);
-                *cp++ = (unsigned char) (temp >> 8);
-                *cp++ = (unsigned char) temp;
+                memcpy (&temp32, cp, sizeof (temp32));
+                *cp++ = (unsigned char) (temp32 >> 24);
+                *cp++ = (unsigned char) (temp32 >> 16);
+                *cp++ = (unsigned char) (temp32 >> 8);
+                *cp++ = (unsigned char) temp32;
                 break;
 
             case 'S':
-                temp = * (int16_t *) cp;
-                *cp++ = (unsigned char) (temp >> 8);
-                *cp++ = (unsigned char) temp;
+                memcpy (&temp16, cp, sizeof (temp16));
+                *cp++ = (unsigned char) (temp16 >> 8);
+                *cp++ = (unsigned char) temp16;
                 break;
 
             default:
