@@ -1,7 +1,7 @@
 ////////////////////////////////////////////////////////////////////////////
 //                           **** WAVPACK ****                            //
 //                  Hybrid Lossless Wavefile Compressor                   //
-//                Copyright (c) 1998 - 2019 David Bryant.                 //
+//                Copyright (c) 1998 - 2024 David Bryant.                 //
 //                          All Rights Reserved.                          //
 //      Distributed under the BSD Software License (see license.txt)      //
 ////////////////////////////////////////////////////////////////////////////
@@ -13,14 +13,26 @@
 
 #include "wavpack.h"
 
-#if defined(_WIN32)
+#if defined(_WIN32) || (defined(__WATCOMC__) && defined(__OS2__))
 #define strdup(x) _strdup(x)
+#ifndef FASTCALL
+#if defined(_M_IX86) || defined(__i386__)
 #define FASTCALL __fastcall
+#else /* !x86: */
+#define FASTCALL
+#endif
+#endif /* FASTCALL */
 #else
 #define FASTCALL
 #endif
 
-#if defined(_WIN32) || \
+#if defined(__WATCOMC__) && defined(OPT_ASM_X86)
+#define ASMCALL __cdecl
+#else
+#define ASMCALL
+#endif
+
+#if defined(_WIN32) || defined(__WATCOMC__) || \
     (defined(BYTE_ORDER) && defined(LITTLE_ENDIAN) && (BYTE_ORDER == LITTLE_ENDIAN)) || \
     (defined(__BYTE_ORDER__) && defined(__ORDER_LITTLE_ENDIAN__) && (__BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__))
 #define BITSTREAM_SHORTS    // use 16-bit "shorts" for reading/writing bitstreams (instead of chars)
@@ -45,6 +57,71 @@ typedef __int8  int8_t;
 #include <stdint.h>
 #endif
 
+// This implements portable multithreading via typedefs and macros for either
+// pthreads or native Windows threads. This is easy since the synchronization
+// constructs we are using (condition variables and mutexes / critical
+// sections) are available on both platforms with similar behavior.
+
+#ifdef ENABLE_THREADS
+
+#ifdef _WIN32
+
+#if defined(_WIN32_WINNT) && _WIN32_WINNT < 0x0600
+#undef _WIN32_WINNT
+#endif
+
+#ifndef _WIN32_WINNT
+#define _WIN32_WINNT 0x0600 /* for CONDITION_VARIABLE & co. */
+#endif
+
+#define WIN32_LEAN_AND_MEAN
+#include <windows.h>
+#include <process.h>
+
+typedef CONDITION_VARIABLE      wp_condvar_t;
+#define wp_condvar_init(x)      InitializeConditionVariable(&x)
+#define wp_condvar_signal(x)    WakeConditionVariable(&x)
+#define wp_condvar_wait(x,y)    SleepConditionVariableCS(&x,&y,INFINITE)
+#define wp_condvar_delete(x)
+
+typedef CRITICAL_SECTION        wp_mutex_t;
+#define wp_mutex_init(x)        InitializeCriticalSection(&x)
+#define wp_mutex_obtain(x)      EnterCriticalSection(&x)
+#define wp_mutex_release(x)     LeaveCriticalSection(&x)
+#define wp_mutex_delete(x)      DeleteCriticalSection(&x)
+
+typedef HANDLE                  wp_thread_t;
+#define wp_thread_create(x,y,z) x=(HANDLE)_beginthreadex(NULL,0,y,z,0,NULL)
+#define wp_thread_join(x)       WaitForSingleObject(x,INFINITE)
+#define wp_thread_delete(x)     CloseHandle(x);
+#define wp_thread_exit(x)       _endthreadex(x);
+
+#else
+
+#include <pthread.h>
+
+typedef pthread_cond_t          wp_condvar_t;
+#define wp_condvar_init(x)      pthread_cond_init(&x,NULL);
+#define wp_condvar_signal(x)    pthread_cond_signal(&x)
+#define wp_condvar_wait(x,y)    pthread_cond_wait(&x,&y)
+#define wp_condvar_delete(x)    pthread_cond_destroy(&x)
+
+typedef pthread_mutex_t         wp_mutex_t;
+#define wp_mutex_init(x)        pthread_mutex_init(&x,NULL);
+#define wp_mutex_obtain(x)      pthread_mutex_lock(&x)
+#define wp_mutex_release(x)     pthread_mutex_unlock(&x)
+#define wp_mutex_delete(x)      pthread_mutex_destroy(&x)
+
+typedef pthread_t               wp_thread_t;
+#define wp_thread_create(x,y,z) do { if (pthread_create(&x,NULL,y,z)) x=0; } while (0)
+#define wp_thread_join(x)       pthread_join(x,NULL)
+#define wp_thread_delete(x)
+#define wp_thread_exit(x)       pthread_exit(x);
+
+#endif
+
+#endif
+
 // Because the C99 specification states that "The order of allocation of
 // bit-fields within a unit (high-order to low-order or low-order to
 // high-order) is implementation-defined" (6.7.2.1), I decided to change
@@ -126,7 +203,6 @@ typedef struct {
 #define CONFIG_MONO_FLAG        4       // not stereo
 #define CONFIG_FLOAT_DATA       0x80    // ieee 32-bit floating point data
 
-#define CONFIG_AUTO_SHAPING     0x4000  // automatic noise shaping
 #define CONFIG_LOSSY_MODE       0x1000000 // obsolete (for information)
 
 /*
@@ -223,16 +299,20 @@ typedef struct {
 } DSDfilters;
 
 typedef struct {
+    const WavpackContext *wpc;
     WavpackHeader wphdr;
     struct words_data w;
+    int stream_index;
 
     unsigned char *blockbuff, *blockend;
     unsigned char *block2buff, *block2end;
-    int32_t *sample_buffer;
+    int32_t *sample_buffer, *pre_sample_buffer;
+    uint32_t num_pre_samples;
+    int discontinuous;
 
     int64_t sample_index;
-    int bits, num_terms, mute_error, joint_stereo, false_stereo, shift;
-    int num_decorrs, num_passes, best_decorr, mask_decorr;
+    int bits, num_terms, mute_error, joint_stereo, false_stereo, shift, lossy_blocks;
+    int num_decorrs, num_passes, best_decorr, mask_decorr, extra_flags;
     uint32_t crc, crc_x, crc_wvx;
     Bitstream wvbits, wvcbits, wvxbits;
     int init_done, wvc_skip;
@@ -240,6 +320,7 @@ typedef struct {
 
     unsigned char int32_sent_bits, int32_zeros, int32_ones, int32_dups;
     unsigned char float_flags, float_shift, float_max_exp, float_norm_exp;
+    unsigned char int32_max_width, float_max_shifted_ones, float_min_shifted_zeros;
 
     struct {
         int32_t shaping_acc [2], shaping_delta [2], error [2];
@@ -274,8 +355,28 @@ typedef struct {
 /////////////////////////////// WavPack Context ///////////////////////////////
 
 // This internal structure holds everything required to encode or decode WavPack
-// files. It is recommended that direct access to this structure be minimized
-// and the provided utilities used instead.
+// files. This is an opaque pointer to clients of libwavpack.
+
+#ifdef ENABLE_THREADS
+
+// Each worker thread owns one of these contexts during its lifetime
+
+typedef enum { Uninit, Ready, Running, Done, Quit } WorkerState;
+
+typedef struct {
+    WavpackStream *wps;
+    WorkerState state;
+    int *workers_ready, *worker_errors;
+    int32_t *outbuf;
+    uint32_t samcnt, offset;
+    int result, free_wps;
+
+    wp_condvar_t *global_cond, worker_cond;
+    wp_mutex_t *mutex;
+    wp_thread_t thread;
+} WorkerInfo;
+
+#endif
 
 struct WavpackContext {
     WavpackConfig config;
@@ -296,11 +397,11 @@ struct WavpackContext {
     int64_t filelen, file2len, filepos, file2pos, total_samples, initial_index;
     uint32_t crc_errors, first_flags;
     int wvc_flag, open_flags, norm_offset, reduced_channels, lossy_blocks, version_five;
-    uint32_t block_samples, ave_block_samples, block_boundary, max_samples, acc_samples, riff_trailer_bytes;
+    uint32_t block_samples, ave_block_samples, block_boundary, max_samples, max_pre_samples, acc_samples, riff_trailer_bytes;
     int riff_header_added, riff_header_created;
     M_Tag m_tag;
 
-    int current_stream, num_streams, max_streams, stream_version;
+    int num_streams, max_streams, stream_version;
     WavpackStream **streams;
     void *stream3;
 
@@ -310,13 +411,22 @@ struct WavpackContext {
     void *decimation_context;
     char file_extension [8];
 
+#ifdef ENABLE_THREADS
+    // these items support multithreaded operations on multichannel streams
+    WorkerInfo *workers;
+    int num_workers, workers_ready, worker_errors;
+    wp_condvar_t global_cond;
+    wp_mutex_t mutex;
+#endif
+
     void (*close_callback)(void *wpc);
     char error_message [80];
 };
 
 //////////////////////// function prototypes and macros //////////////////////
 
-#define CLEAR(destin) memset (&destin, 0, sizeof (destin));
+#define CLEAR(destin) memset (&destin, 0, sizeof (destin))
+#define CLEARA(destin) memset (destin, 0, sizeof (destin))  /* for arrays */
 
 //////////////////////////////// decorrelation //////////////////////////////
 // modules: pack.c, unpack.c, unpack_floats.c, extra1.c, extra2.c
@@ -369,33 +479,33 @@ struct WavpackContext {
         weight = (weight ^ s) - s; \
     }
 
-void pack_init (WavpackContext *wpc);
-int pack_block (WavpackContext *wpc, int32_t *buffer);
-void send_general_metadata (WavpackContext *wpc);
+void pack_init (WavpackStream *wps);
+int pack_block (WavpackStream *wps, int32_t *buffer);
+void send_general_metadata (WavpackStream *wps);
+void send_pending_metadata (WavpackStream *wps);
 void free_metadata (WavpackMetadata *wpmd);
 int copy_metadata (WavpackMetadata *wpmd, unsigned char *buffer_start, unsigned char *buffer_end);
 double WavpackGetEncodedNoise (WavpackContext *wpc, double *peak);
-int unpack_init (WavpackContext *wpc);
+int unpack_init (WavpackContext *wpc, int stream);
 int read_decorr_terms (WavpackStream *wps, WavpackMetadata *wpmd);
 int read_decorr_weights (WavpackStream *wps, WavpackMetadata *wpmd);
 int read_decorr_samples (WavpackStream *wps, WavpackMetadata *wpmd);
 int read_shaping_info (WavpackStream *wps, WavpackMetadata *wpmd);
-int32_t unpack_samples (WavpackContext *wpc, int32_t *buffer, uint32_t sample_count);
-int check_crc_error (WavpackContext *wpc);
+int32_t unpack_samples (WavpackStream *wps, int32_t *buffer, uint32_t sample_count);
 int scan_float_data (WavpackStream *wps, f32 *values, int32_t num_values);
 void send_float_data (WavpackStream *wps, f32 *values, int32_t num_values);
 void float_values (WavpackStream *wps, int32_t *values, int32_t num_values);
-void dynamic_noise_shaping (WavpackContext *wpc, int32_t *buffer, int shortening_allowed);
-void execute_stereo (WavpackContext *wpc, int32_t *samples, int no_history, int do_samples);
-void execute_mono (WavpackContext *wpc, int32_t *samples, int no_history, int do_samples);
+void dynamic_noise_shaping (WavpackStream *wps, const int32_t *buffer, int shortening_allowed);
+void execute_stereo (WavpackStream *wps, int32_t *samples, int no_history, int do_samples);
+void execute_mono (WavpackStream *wps, int32_t *samples, int no_history, int do_samples);
 
 ////////////////////////// DSD related (including decimation) //////////////////////////
 // modules: pack_dsd.c unpack_dsd.c
 
-void pack_dsd_init (WavpackContext *wpc);
-int pack_dsd_block (WavpackContext *wpc, int32_t *buffer);
-int init_dsd_block (WavpackContext *wpc, WavpackMetadata *wpmd);
-int32_t unpack_dsd_samples (WavpackContext *wpc, int32_t *buffer, uint32_t sample_count);
+void pack_dsd_init (WavpackStream *wps);
+int pack_dsd_block (WavpackStream *wps, int32_t *buffer);
+int init_dsd_block (WavpackStream *wps, WavpackMetadata *wpmd);
+int32_t unpack_dsd_samples (WavpackStream *wps, int32_t *buffer, uint32_t sample_count);
 
 void *decimate_dsd_init (int num_channels);
 void decimate_dsd_reset (void *decimate_context);
@@ -404,7 +514,8 @@ void decimate_dsd_destroy (void *decimate_context);
 
 ///////////////////////////////// CPU feature detection ////////////////////////////////
 
-int unpack_cpu_has_feature_x86 (int findex), pack_cpu_has_feature_x86 (int findex);
+int ASMCALL unpack_cpu_has_feature_x86 (int findex);
+int ASMCALL pack_cpu_has_feature_x86 (int findex);
 
 #define CPU_FEATURE_MMX     23
 
@@ -514,6 +625,14 @@ uint32_t bs_close_read (Bitstream *bs);
 
 #ifdef HAVE___BUILTIN_CLZ
 #define count_bits(av) ((av) ? 32 - __builtin_clz (av) : 0)
+#elif defined (__WATCOMC__) && defined(__386__)
+extern __inline int _bsr_watcom(uint32_t);
+#pragma aux _bsr_watcom = \
+  "bsr eax, eax" \
+  parm [eax] nomemory \
+  value [eax] \
+  modify exact [eax] nomemory;
+#define count_bits(av) ((av) ? _bsr_watcom((av)) + 1 : 0)
 #elif defined (_WIN64)
 static __inline int count_bits (uint32_t av) { unsigned long res; return _BitScanReverse (&res, av) ? (int)(res + 1) : 0; }
 #else
@@ -558,7 +677,7 @@ int FASTCALL wp_log2 (uint32_t avalue);
 #define LOG2BUFFER log2buffer
 #endif
 
-uint32_t LOG2BUFFER (int32_t *samples, uint32_t num_samples, int limit);
+uint32_t ASMCALL LOG2BUFFER (int32_t *samples, uint32_t num_samples, int limit);
 
 signed char store_weight (int weight);
 int restore_weight (signed char weight);
@@ -604,7 +723,7 @@ int WavpackGetMD5Sum (WavpackContext *wpc, unsigned char data [16]);
 
 int WavpackVerifySingleBlock (unsigned char *buffer, int verify_checksum);
 uint32_t read_next_header (WavpackStreamReader64 *reader, void *id, WavpackHeader *wphdr);
-int read_wvc_block (WavpackContext *wpc);
+int read_wvc_block (WavpackContext *wpc, int stream);
 
 /////////////////////////// high-level packing API and support ////////////////////////////
 // modules: pack_utils.c, pack_floats.c
@@ -657,6 +776,7 @@ void WavpackBigEndianToNative (void *data, char *format);
 void WavpackNativeToBigEndian (void *data, char *format);
 
 void install_close_callback (WavpackContext *wpc, void cb_func (void *wpc));
+void free_single_stream (WavpackStream *wps);
 void free_dsd_tables (WavpackStream *wps);
 void free_streams (WavpackContext *wpc);
 
